GSM_ENCODING = 'gsm'
GSM_MAX_SEGMENT_SIZE = 160
GSM_MAX_CONCAT_SEGMENT_SIZE = 153
UCS2_ENCODING = 'ucs2'
UCS2_MAX_SEGMENT_SIZE = 140
UCS2_MAX_CONCAT_SEGMENT_SIZE = 134
MMS_MAX_SEGMENT_SIZE = 1600
MMS_MAX_CONCAT_SEGMENT_SIZE = 1600
